{{
  HSPID
High speed pulse interval and duration measuring program.  Written 14/2/2013 T:=22:03
First thought about how fast Prop could measure pulse intervals and durations when
was writing code to interface to Polar HR monitor as wanted to ensure that every
pulse counted was 15 msec in duration (they were).  Then after a few mental calculations
it became apparent that one can measure pulses and intervals to very high precision.

When run at an 80 MHz clock speed, the timer loop in this program has a cycle time of
2 microseconds which is amazingly fast for something that's done purely in code.  It can
be speeded up a bit if one uses either the c or z flag instead of variable TState, but
2 microseconds was fast enough for my application and so will try to optimize it some
other day.

Right now only minimal spin code can run if there's a lot of data being produced.  It's
left as an exercise for the user to include FIFO routines in hub RAM to deal with data
bursts.

At present, this program measures pulses on pin4 - change the value of DataMask in PASM
code to chose another pin.  All pins are assume to be set as input when the program starts
up.  Initialization of the code in a cog is performed by
  SetupTimer(@Tbegin, @PulseTime)

This routine loads the code into a new cog and initialized the cog with the address of the
parameter area used.  To get a value back, use the other routine in the object:
  retcode :=ReadTimer(Ptime, Pdur)
  where:
    retcode - is 0 if data returned, -1 otherwise
    Ptime   - long variable holding the timer count of start of pulse
    Pdur    - duration of the pulse in units of 2 microseconds

Most of the programs time will be spent in a tight loop calling ReadTimer and copying out
data to an array when it does arrive.

Cog access to hub RAM takes advantage of the fact that, in a Spin object, variables defined
in a VAR block are all at sequential addresses.  The structure of the HSPID Spin parameter area
is:
  PulseTime  long timestamp of the start of the pulse in units of 2 microseconds
  PulseDur   long Duration of the pulse in 2 microsecond ticks
  PWritten   long Number of pulse events written to hub RAM
  PRead      long Number of pulses read by spin

Reason for using the PWritten and PRead variables is that it frees one from having to use
Mutexes.  PWritten is only written by cog and PRead is only written by Spin.  Every time
a new event is posted, also increment PWritten.  When an event is read by Spin, it then
increments PRead.  If these two variables are equal, it means no new events have occurred.
If PWritten - Pread >1 then it means that events have been missed (you'll have to write your
own routines to do this as the spin code for this project took longer than the PASM).

If you need long term timing, then you'll have to add a second timer long as the current
timer overflows in 8589 seconds at 2 MHz.

Main reason I wrote this program was to reverse engineer the HeartLink monitor to serial port
interface.  The company Heartlink is now defunct and the only program they had to read the
monitors ran in DOS and used some bizarre bitbanging hack with the serial port to read and
initialize the monitors.  I'm glad that the Propeller is fast enough to do this for me as I
don't have the budget for a logic analyzer.  Also, my last DOS laptop whose serial port was
compatible with the HL monitors died last year.

Pencil and paper timing of the program indicates that it takes: (actually more)
  32 clocks/loop if no pulse is encountered
  44 clocks for discovering the start of a pulse
  44 clocks to update pulse duration while pulse still high
 105-120 clocks to output the pulse.  Not much room for improvement there as there's that
         hub uncertainty that exists although there's stil room for one more instruction
         after the second hum write operation if one wants to maximize overlapped processing.
Note: after debugging routine, had to add extra code so tack on another 10-20 cycles for the
whole routine.  Just made it under that 2 microsecond/cycle barrier

(C) Dr. Boris Gimbarzevsky 2013
This program is licensed under the GPL.  The only thing I ask is that you mention where you
got the program from when you've hacked it to do what you want.
}}
CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  'bits = 10  


VAR
    long              PulseTime         
    long              PulseDur          
    long              PWritten          
    long              PRead             

PUB SetupTimer   : Tset
    tset :=cognew(@TBegin, @PulseTime)
    return

PUB GetPulseTime : value
    value := PulseTime
    return
PUB GetPulseDur  : value
    value := PulseDur
    return
PUB GetPWritten : value
    value := PWritten
    return
PUB GetPRead : value
    value := PRead
    return
PUB IncPRead
    PRead++
    Return
'
' I really really am starting to hate spin.  The routine below doesn't work and so have
' to return variables one at a time.  PASM is so so much easier and logical
'
PUB ReadTimer(PT, PD) : value
    value :=-1
    if (PWritten > PRead)
      PT := PulseTime
      PD := PulseDur
      PRead++
      value := 0
    return

DAT
                  org

TBegin            mov     HubAddr, par                  ' get add of hub comm area
                  xor     THigh, THigh                  ' set THigh to 0
                  xor     TState, TState  wz            'pulse state var = 0
                  muxz    dira, #DataMask               ' set bit4 to input                  ' 
                  mov     TmTime, cnt
                  add     TmTime, TimeTick
Timer             cmp     TmTime, cnt  wz, wc           ' see if cnt >= to TmTime
          if_nc   jmp     #Timer                        ' no, jump back
                  add     TmTime, TimeTick              ' setup next time benchmark to reach
                  add     THigh, #1                     ' increment our internal 2 microsec timer
                  mov     Data, ina                     ' get all of the data lines
                  and     Data, DataMask                ' leave behind only the bit we're interested in
                  tjnz    Tstate, #TS1                  ' tstate = 1, go to its code
                  tjz     Data, #Timer                  ' state = 0 and no pulse, try again next clock
                  '
                  ' State 0 just gets start of pulse
                  '
                  mov     TPulse, THigh                 ' time of pulse start
                  mov     PDur, #1                      ' at least 1 clock long now
                  add     TState, #1                    ' now state machine is in state 1
                  jmp     #Timer
                  '
                  ' State 1 either increments duration or outputs pulse if sample after trailing edge
                  '
                  '
TS1               tjz     Data, #TPend                  ' if Data = 0, time to wrap up
                  add     PDur, #1                      ' update PDur if still in pulse
                  jmp     #Timer
                  '
                  ' Here output data to hub RAM - takes longest to execute
                  '
TPend             wrlong  TPulse, HubAddr               ' first write the start time of pulse
                  add     HubAddr, #4                   ' make HubAddr point to next variable
                  ' can execute another instruction here for free
                  wrlong  PDur, HubAddr                 ' Send pulse duration to hub
                  add     PWrite, #1                    ' let hub know there's data
                  add     HubAddr, #4                   ' now we've used those 8 clocks
                  wrlong  PWrite, HubAddr
                  sub     HubAddr, #8                   ' return HubAddr to its pristine condition
                  xor     PDur, Pdur
                  xor     TState, TState
                  xor     TPulse, TPulse                ' zero out variables for next pass
                  jmp     #Timer
'
' Here are the cog variables
'
TmTime            long          0
TimeTick          long        160                       ' 2 microseconds
THigh             long          0                       ' Running time kept here
HubAddr           long          0
DataMask          long         $10                       ' pin4 
TPulse            long          0
PDur              long          0
TState            long          0
PWrite            long          0
Data              long          0
